// Fichier libre de droits

#ifndef SOLITAIRE_H
#define SOLITAIRE_H

#include <iostream>
#include <sstream>
#include <list>
#include <vector>
#include <set>
#include <map>
#include "ContenuFichier.h"
#include "EnsembleDeCases.h"
#include "Mouvement.h"

template<class ImplementationInterne>
struct Solitaire {
	Solitaire(const ContenuFichier &);
	bool solve(bool);
	void explore(std::ostream &);
	void afficheSolution(std::ostream &) const;
	void afficheExploration(std::ostream &) const;
	void afficheDictionnaires(std::ostream &) const;
private:
	bool essaye(int n);
	void display(std::ostream &, const EnsembleDeCases<ImplementationInterne> &) const;
	static void afficheAccord(std::ostream &, size_t, const char *);

	EnsembleDeCases<ImplementationInterne> plateau;
	EnsembleDeCases<ImplementationInterne> pions;
	EnsembleDeCases<ImplementationInterne> objectif;

	int tailleFinale;

	bool avecMemoire;
	bool solved;

	// Pour la recherche rcursive
	std::vector<std::set<EnsembleDeCases<ImplementationInterne> > > dejaVu;
	std::vector<EnsembleDeCases<ImplementationInterne> > solution;

	// Pour l'exploration complte
	std::vector<size_t> nombreDeConfigurations;
	std::vector<size_t> nombreDImpasses;
	std::vector<size_t> nombreDeCheminsSansIssue;
	typename std::map<EnsembleDeCases<ImplementationInterne>, size_t> etatsActuels;
};

template<class ImplementationInterne>
Solitaire<ImplementationInterne>::Solitaire(const ContenuFichier &plateauDeJeu) {
	plateau.init(plateauDeJeu.hexagonal, plateauDeJeu.ligneMaximum);
	pions.init(plateauDeJeu.hexagonal, plateauDeJeu.ligneMaximum);
	objectif.init(plateauDeJeu.hexagonal, plateauDeJeu.ligneMaximum);
	for (std::list<ContenuFichier::Point>::const_iterator i = plateauDeJeu.casesOccupees.begin(); i != plateauDeJeu.casesOccupees.end(); ++i) {
		typename ImplementationInterne::Point p(i->ligne, i->colonne);
		plateau.ajoute(p);
		pions.ajoute(p);
	}
	for (std::list<ContenuFichier::Point>::const_iterator i = plateauDeJeu.casesLibres.begin(); i != plateauDeJeu.casesLibres.end(); ++i) {
		typename ImplementationInterne::Point p(i->ligne, i->colonne);
		plateau.ajoute(p);
		objectif.ajoute(p);
	}
	plateau.initListeMouvements();
	tailleFinale = objectif.taille();
	solved = false;
	avecMemoire = false;
}

template<class ImplementationInterne>
bool Solitaire<ImplementationInterne>::essaye(int n) {
	if (n == tailleFinale) {
		if (objectif == pions) {
			solution[n] = pions;
			return true;
		}
		return false;
	}
	if (avecMemoire) {
		if (0 < dejaVu[n].count(pions)) {
			return false;
		}
		dejaVu[n].insert(pions);
	}
	for (typename std::list<Mouvement<typename ImplementationInterne::Point> >::const_iterator i = plateau.listeMouvements.begin(); i != plateau.listeMouvements.end(); ++i) {
		if (pions.appliqueMouvement(*i)) {
			if (essaye(n - 1)) {
				pions.retireMouvement(*i);
				solution[n] = pions;
				return true;
			}
			pions.retireMouvement(*i);
		}
	}
	return false;
}

template<class ImplementationInterne>
bool Solitaire<ImplementationInterne>::solve(bool utiliserDictionnaire) {
	avecMemoire = utiliserDictionnaire;
	solved = false;
	dejaVu.clear();
	int n = pions.taille();
	dejaVu.resize(n + 1);
	solution.resize(n + 1);
	solved = essaye(n);
	return solved;
}

template<class ImplementationInterne>
void Solitaire<ImplementationInterne>::afficheSolution(std::ostream &os) const {
	if (! solved) {
		return;
	}
	for (int i = pions.taille(); i >= tailleFinale; --i) {
		display(os, solution[i]);
	}
}

template<class ImplementationInterne>
void Solitaire<ImplementationInterne>::afficheDictionnaires(std::ostream &os) const {
	if (! solved) {
		return;
	}
	for (int i = pions.taille(); i >= tailleFinale; --i) {
		os << "Dictionnaire[" << i << "] : " << dejaVu[i].size() << '\n';
	}
}

template<class ImplementationInterne>
void Solitaire<ImplementationInterne>::display(std::ostream &os, const EnsembleDeCases<ImplementationInterne> &etat) const {
	std::stringstream affichagePlateau;
	plateau.display(affichagePlateau, ".");
	std::stringstream casesOccupes;
	casesOccupes << etat;
	std::string ligneFond;
	std::string ligneCasesOccupes;
	while ( getline(affichagePlateau, ligneFond) ) {
		if (getline(casesOccupes, ligneCasesOccupes) ) {
			for (unsigned int i = 0; i < ligneCasesOccupes.size(); ++i) {
				// Devrait bien tre la ligne la plus courte...
				// Et sinon, le .at() doit lancer une exception.
				if (' ' != ligneCasesOccupes.at(i)) {
					ligneFond.at(i) = ligneCasesOccupes.at(i);
				}
			}
		}
		os << ligneFond << '\n';
	}
	os << '\n';
}

template<class ImplementationInterne>
void Solitaire<ImplementationInterne>::explore(std::ostream &os) {
	typename std::map<EnsembleDeCases<ImplementationInterne>, size_t> etatsSuivants;
	solved = false;
	etatsActuels.clear();
	etatsActuels[pions] = 1;
	int n = pions.taille();
	nombreDeConfigurations.resize(n + 1, 0);
	nombreDImpasses.resize(n + 1, 0);
	nombreDeCheminsSansIssue.resize(n + 1, 0);
	while (n > tailleFinale) {
		nombreDeConfigurations[n] = etatsActuels.size();
		os << "Avec ";
		afficheAccord(os, n, " pion");
		os << " : ";
		afficheAccord(os, nombreDeConfigurations[n], " configuration");
		for (typename std::map<EnsembleDeCases<ImplementationInterne>, size_t>::const_iterator mapIterator = etatsActuels.begin(); mapIterator != etatsActuels.end(); ++mapIterator) {
			bool impasse = true;
			EnsembleDeCases<ImplementationInterne> etatCourant(mapIterator->first);
			for (typename std::list<Mouvement<typename ImplementationInterne::Point> >::const_iterator i = plateau.listeMouvements.begin(); i != plateau.listeMouvements.end(); ++i) {
				if (etatCourant.appliqueMouvement(*i)) {
					etatsSuivants[etatCourant] += mapIterator->second;
					etatCourant.retireMouvement(*i);
					impasse = false;
				}
			}
			if (impasse) {
				nombreDImpasses[n] += 1;
				nombreDeCheminsSansIssue[n] += mapIterator->second;
			}
		}
		os << " dont " << nombreDImpasses[n] << " sans suite possible\n";
		etatsActuels.clear();
		etatsActuels.swap(etatsSuivants);
		n -= 1;
	}
	nombreDeConfigurations[n] = etatsActuels.size();
	solved = (etatsActuels.count(objectif) > 0);
	afficheAccord(os, nombreDeConfigurations[tailleFinale], " configuration");
	os << " terminale";
	if (nombreDeConfigurations[tailleFinale] > 1) {
		os << "s";
	}
	if (solved) {
		os << ", dont la solution.\n";
	} else {
		os << ", sans solution.\n";
	}
}

template<class ImplementationInterne>
void Solitaire<ImplementationInterne>::afficheAccord(std::ostream &os, size_t nombre, const char *machin) {
	os << nombre << machin;
	if (nombre > 1) {
		os << "s";
	}
}

template<class ImplementationInterne>
void Solitaire<ImplementationInterne>::afficheExploration(std::ostream &os) const {
	size_t blocages = 0;
	size_t nombreSolutions = 0;
	size_t nombrePresqueSolutions = 0;
	for (int n = pions.taille(); n > tailleFinale; --n) {
		blocages += nombreDeCheminsSansIssue[n];
	}
	for (typename std::map<EnsembleDeCases<ImplementationInterne>, size_t>::const_iterator mapIterator = etatsActuels.begin(); mapIterator != etatsActuels.end(); ++mapIterator) {
		if (mapIterator->first == objectif) {
			nombreSolutions += mapIterator->second;
		} else {
			nombrePresqueSolutions += mapIterator->second;
		}
	}
	afficheAccord(os, blocages, " chemin");
	os << " sans issue\n";
	afficheAccord(os, nombrePresqueSolutions, " chemin");
	os << " chemins arrivant presque au bout (mauvaise configuration finale)\n";
	afficheAccord(os, nombreSolutions, " chemin");
	os << " chemins arrivant pour la solution" << std::endl;
}

#endif // SOLITAIRE_H
